from datetime import datetime, timedelta, timezone
from decimal import Decimal
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import call, sentinel

from fred_api.registry.registrar.registrar_common_types_pb2 import Credit, RegistrarHandle
from fred_api.registry.registrar.registrar_credit_types_pb2 import RegistrarCreditReply, RegistrarCreditRequest
from fred_api.registry.registrar.registrar_info_types_pb2 import RegistrarInfoReply, RegistrarInfoRequest
from fred_api.registry.registrar.registrar_list_types_pb2 import RegistrarListReply
from fred_api.registry.registrar.registrar_zone_access_types_pb2 import (RegistrarZoneAccessReply,
                                                                         RegistrarZoneAccessRequest)
from fred_api.registry.registrar.service_admin_grpc_pb2 import (AddRegistrarZoneAccessReply,
                                                                AddRegistrarZoneAccessRequest, CreateRegistrarReply,
                                                                CreateRegistrarRequest, DeleteRegistrarZoneAccessReply,
                                                                DeleteRegistrarZoneAccessRequest, UpdateRegistrarReply,
                                                                UpdateRegistrarRequest, UpdateRegistrarZoneAccessReply,
                                                                UpdateRegistrarZoneAccessRequest)
from freezegun import freeze_time
from frgal.utils import AsyncTestClientMixin, make_awaitable
from google.protobuf.empty_pb2 import Empty

from regal.common import Address
from regal.registrar import Registrar, RegistrarAdminClient, RegistrarClient, RegistrarDecoder, ZoneAccess


class TestRegistrarClient(AsyncTestClientMixin, RegistrarClient):
    """Test RegistrarClient."""


class TestRegistrarAdminClient(AsyncTestClientMixin, RegistrarAdminClient):
    """Test RegistrarAdminClient."""


@freeze_time("2000-01-01T00:00:00Z")
class ZoneAccessTest(TestCase):
    def test_is_active(self):
        now = datetime.now(tz=timezone.utc)
        data = (
            (ZoneAccess(zone="cz"), True),
            (ZoneAccess(zone="cz", valid_from=now), True),
            (ZoneAccess(zone="cz", valid_from=now - timedelta(days=1)), True),
            (ZoneAccess(zone="cz", valid_from=now + timedelta(days=1)), False),
            (ZoneAccess(zone="cz", valid_to=now), False),
            (ZoneAccess(zone="cz", valid_to=now - timedelta(days=1)), False),
            (ZoneAccess(zone="cz", valid_to=now + timedelta(days=1)), True),
            (ZoneAccess(zone="cz", valid_from=now - timedelta(days=1), valid_to=now + timedelta(days=1)), True),
            (ZoneAccess(zone="cz", valid_from=now - timedelta(days=2), valid_to=now - timedelta(days=1)), False),
            (ZoneAccess(zone="cz", valid_from=now + timedelta(days=1), valid_to=now + timedelta(days=2)), False),
        )
        for access, is_valid in data:
            with self.subTest(access=access, is_valid=is_valid):
                self.assertIs(access.is_active(), is_valid)


class RegistrarDecoderTest(TestCase):

    def setUp(self):
        self.decoder = RegistrarDecoder()

    def test_decode_credit(self):
        self.assertEqual(self.decoder.decode(Credit(decimal_value='42.19')), Decimal('42.19'))
        self.assertIsNone(self.decoder.decode(Credit()))


class RegistrarClientTest(IsolatedAsyncioTestCase):
    """Test RegistrarClient's methods."""

    def setUp(self):
        self.client = TestRegistrarClient(netloc=sentinel.netloc)

    async def test_get_registrar_handles(self):
        registrar_handles = RegistrarListReply()
        registrar_handles.data.registrar_handles.extend([
            RegistrarHandle(value='REG-SW'),
            RegistrarHandle(value='REG-ST'),
        ])
        self.client.mock.return_value = registrar_handles

        self.assertEqual(await self.client.get_registrar_handles(), ['REG-SW', 'REG-ST'])

        self.client.mock.assert_called_once_with(
            Empty(),
            method='/Fred.Registry.Api.Registrar.Registrar/get_registrar_handles',
            timeout=None,
        )

    async def test_get_registrar_info(self):
        registrar_info = RegistrarInfoReply()
        registrar_info.data.registrar_handle.value = 'REG-SW'
        registrar_info.data.name = 'Star Wars Inc.'
        registrar_info.data.place.street[:] = ['Narrow street 12']
        registrar_info.data.place.city = 'Cloud City'
        registrar_info.data.place.postal_code.value = '12345'
        registrar_info.data.place.country_code.value = 'GE'
        self.client.mock.return_value = registrar_info

        reply = await self.client.get_registrar_info('REG-SW')
        self.assertEqual(reply, Registrar(
            registrar_handle='REG-SW',
            name='Star Wars Inc.',
            place=Address(
                street=['Narrow street 12'],
                city='Cloud City',
                postal_code='12345',
                country_code='GE',
            ),
        ))

        request = RegistrarInfoRequest()
        request.registrar_handle.value = 'REG-SW'
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Registrar.Registrar/get_registrar_info',
            timeout=None,
        )

    async def test_get_registrar_credit(self):
        registrar_credit = RegistrarCreditReply()
        registrar_credit.data.credit_by_zone['cz'].decimal_value = '42000.40'
        registrar_credit.data.credit_by_zone['0.2.4.e164.arpa'].decimal_value = '0.00'
        self.client.mock.return_value = registrar_credit

        self.assertEqual(await self.client.get_registrar_credit('REG-SW'), {
            'cz': Decimal('42000.40'),
            '0.2.4.e164.arpa': Decimal('0.00'),
        })

        request = RegistrarCreditRequest()
        request.registrar_handle.value = 'REG-SW'
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Registrar.Registrar/get_registrar_credit',
            timeout=None,
        )

    async def test_get_registrar_zone_access(self):
        registrar_zone_access = RegistrarZoneAccessReply()
        registrar_zone_access.data.access_by_zone['cz'].has_access_now = True
        registrar_zone_access.data.access_by_zone['0.2.4.e164.arpa'].has_access_now = False
        self.client.mock.return_value = registrar_zone_access

        self.assertEqual(await self.client.get_registrar_zone_access('REG-SW'), {
            'cz': True,
            '0.2.4.e164.arpa': False,
        })

        request = RegistrarZoneAccessRequest()
        request.registrar_handle.value = 'REG-SW'
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Registrar.Registrar/get_registrar_zone_access',
            timeout=None,

        )

    async def test_get_registrar_zone_access_history(self):
        registrar_zone_access = RegistrarZoneAccessReply()
        registrar_zone_access.data.access_by_zone['cz'].has_access_now = True
        history = registrar_zone_access.data.access_by_zone['cz'].history.add()
        history.valid_from.FromDatetime(datetime(2020, 1, 1, tzinfo=timezone.utc))
        registrar_zone_access.data.access_by_zone['0.2.4.e164.arpa'].has_access_now = False
        history = registrar_zone_access.data.access_by_zone['0.2.4.e164.arpa'].history.add()
        history.valid_from.FromDatetime(datetime(2020, 1, 1, tzinfo=timezone.utc))
        history.valid_to.FromDatetime(datetime(2021, 1, 1, tzinfo=timezone.utc))
        self.client.mock.return_value = registrar_zone_access

        self.assertEqual(await self.client.get_registrar_zone_access_history('REG-SW'), {
            'cz': [ZoneAccess(
                zone='cz',
                valid_from=datetime(2020, 1, 1, tzinfo=timezone.utc),
                valid_to=None,
            )],
            '0.2.4.e164.arpa': [ZoneAccess(
                zone='0.2.4.e164.arpa',
                valid_from=datetime(2020, 1, 1, tzinfo=timezone.utc),
                valid_to=datetime(2021, 1, 1, tzinfo=timezone.utc),
            )],
        })

        request = RegistrarZoneAccessRequest()
        request.registrar_handle.value = 'REG-SW'
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Registrar.Registrar/get_registrar_zone_access',
            timeout=None,

        )


class RegistrarAdminClientTest(IsolatedAsyncioTestCase):
    """Test RegistrarAdminClient's methods."""

    def setUp(self):
        self.client = TestRegistrarAdminClient(netloc=sentinel.netloc)
        self.now = datetime(2022, 1, 1, tzinfo=timezone.utc)
        self.tomorrow = self.now + timedelta(days=1)

    async def test_create_registrar_minimal(self):
        registrar = Registrar(registrar_handle='HOLLY', name='Holly', organization='JMC')
        reply = CreateRegistrarReply()
        reply.data.registrar_handle.value = 'HOLLY'
        self.client.mock.return_value = reply

        self.assertIsNone(await self.client.create_registrar(registrar))

        request = CreateRegistrarRequest()
        request.registrar_handle.value = 'HOLLY'
        request.name = 'Holly'
        request.organization = 'JMC'
        calls = [call(request,
                      method='/Fred.Registry.Api.Registrar.Admin/create_registrar',
                      timeout=None)]
        self.assertEqual(self.client.mock.mock_calls, calls)

    async def test_create_registrar(self):
        place = Address(street=['Deck 16'], city='Red Dwarf', state_or_province='Deep Space', postal_code='JMC',
                        country_code='JU')
        registrar = Registrar(
            registrar_handle='HOLLY', name='Holly', organization='JMC', place=place, telephone='+1.234', fax='+9.876',
            emails=['holly@example.org'], url='jmc.example.org', company_registration_number='JMC42',
            vat_identification_number='JU42', is_system_registrar=True, variable_symbol='42', payment_memo_regex='.*',
            is_vat_payer=True)
        reply = CreateRegistrarReply()
        reply.data.registrar_handle.value = 'HOLLY'
        self.client.mock.return_value = reply

        self.assertIsNone(await self.client.create_registrar(registrar))

        request = CreateRegistrarRequest()
        request.registrar_handle.value = 'HOLLY'
        request.name = 'Holly'
        request.organization = 'JMC'
        request.place.street.append('Deck 16')
        request.place.city = 'Red Dwarf'
        request.place.state_or_province = 'Deep Space'
        request.place.postal_code.value = 'JMC'
        request.place.country_code.value = 'JU'
        request.telephone.value = '+1.234'
        request.fax.value = '+9.876'
        request.emails.add(value='holly@example.org')
        request.url.value = 'jmc.example.org'
        request.is_system_registrar = True
        request.company_registration_number.value = 'JMC42'
        request.vat_identification_number.value = 'JU42'
        request.variable_symbol = '42'
        request.payment_memo_regex = '.*'
        request.is_vat_payer = True
        calls = [call(request,
                      method='/Fred.Registry.Api.Registrar.Admin/create_registrar',
                      timeout=None)]
        self.assertEqual(self.client.mock.mock_calls, calls)

    async def test_update_registrar_minimal(self):
        reply = UpdateRegistrarReply()
        reply.data.registrar_handle.value = "HOLLY"
        self.client.mock.return_value = make_awaitable(reply)

        self.assertIsNone(await self.client.update_registrar("HOLLY"))

        request = UpdateRegistrarRequest()
        request.registrar_handle.value = "HOLLY"
        self.assertEqual(self.client.mock.mock_calls, [
            call(request, method="/Fred.Registry.Api.Registrar.Admin/update_registrar", timeout=None),
        ])

    async def test_update_registrar_minimal_address(self):
        address = Address(street=["Deck 5"], city="Red Dwarf")
        reply = UpdateRegistrarReply()
        reply.data.registrar_handle.value = "HOLLY"
        self.client.mock.return_value = make_awaitable(reply)

        self.assertIsNone(await self.client.update_registrar("HOLLY", place=address))

        request = UpdateRegistrarRequest()
        request.registrar_handle.value = "HOLLY"
        request.set_place.street[:] = ["Deck 5"]
        request.set_place.city = "Red Dwarf"
        self.assertEqual(self.client.mock.mock_calls, [
            call(request, method="/Fred.Registry.Api.Registrar.Admin/update_registrar", timeout=None),
        ])

    async def test_update_registrar(self):
        address = Address(street=["Deck 5"], city="Red Dwarf", postal_code="123", country_code="JMC")
        reply = UpdateRegistrarReply()
        reply.data.registrar_handle.value = "HOLLY"
        self.client.mock.return_value = make_awaitable(reply)

        self.assertIsNone(await self.client.update_registrar(
            "HOLLY", name="Holly", organization="Jupiter Mining Corp.", place=address, telephone="+420.111222333",
            fax="+420.444555666", emails=["holly@example.com"], url="http://example.com", is_system_registrar=True,
            company_registration_number="12345", vat_identification_number="67890", variable_symbol="VAR",
            payment_memo_regex="REGEX", is_vat_payer=True, is_internal=False,
        ))

        request = UpdateRegistrarRequest()
        request.registrar_handle.value = "HOLLY"
        request.set_name = "Holly"
        request.set_organization = "Jupiter Mining Corp."
        request.set_place.street[:] = ["Deck 5"]
        request.set_place.city = "Red Dwarf"
        request.set_place.postal_code.value = "123"
        request.set_place.country_code.value = "JMC"
        request.set_telephone.value = "+420.111222333"
        request.set_fax.value = "+420.444555666"
        request.set_emails.values.add(value="holly@example.com")
        request.set_url.value = "http://example.com"
        request.set_is_system_registrar = True
        request.set_company_registration_number.value = "12345"
        request.set_vat_identification_number.value = "67890"
        request.set_variable_symbol = "VAR"
        request.set_payment_memo_regex = "REGEX"
        request.set_is_vat_payer = True
        request.set_is_internal = False
        self.assertEqual(self.client.mock.mock_calls, [
            call(request, method="/Fred.Registry.Api.Registrar.Admin/update_registrar", timeout=None),
        ])

    def _get_add_zone_access_request(self, registrar="REG-SW", zone="cz", valid_from=None, valid_to=None):
        request = AddRegistrarZoneAccessRequest()
        request.registrar_handle.value = registrar
        request.zone = zone
        if valid_from:
            request.valid_from.FromDatetime(valid_from)
        if valid_to:
            request.valid_to.FromDatetime(valid_to)
        return request

    def _get_zone_access_reply(
        self, reply, registrar="REG-SW", zone="cz", valid_from=datetime(2022, 1, 1, tzinfo=timezone.utc), valid_to=None,
    ):
        reply.data.registrar_handle.value = "REG-SW"
        reply.data.zone = "cz"
        reply.data.valid_from.FromDatetime(valid_from)
        if valid_to:
            reply.data.valid_to.FromDatetime(valid_to)
        return reply

    def _get_add_zone_access_reply(self, *args, **kwargs):
        return self._get_zone_access_reply(AddRegistrarZoneAccessReply(), *args, **kwargs)

    def _get_update_zone_access_reply(self, *args, **kwargs):
        return self._get_zone_access_reply(UpdateRegistrarZoneAccessReply(), *args, **kwargs)

    async def test_add_zone_access_without_limits(self):
        self.client.mock.return_value = make_awaitable(self._get_add_zone_access_reply())

        self.assertEqual(
            await self.client.add_zone_access("REG-SW", ZoneAccess(zone="cz")),
            ZoneAccess(zone="cz", valid_from=self.now),
        )

        self.client.mock.assert_called_once_with(
            self._get_add_zone_access_request(),
            method="/Fred.Registry.Api.Registrar.Admin/add_registrar_zone_access",
            timeout=None,
        )

    async def test_add_zone_access_with_limits(self):
        self.client.mock.return_value = make_awaitable(self._get_add_zone_access_reply(valid_to=self.tomorrow))

        zone_access = ZoneAccess(zone="cz", valid_from=self.now, valid_to=self.tomorrow)
        self.assertEqual(await self.client.add_zone_access("REG-SW", zone_access), zone_access)

        self.client.mock.assert_called_once_with(
            self._get_add_zone_access_request(valid_from=self.now, valid_to=self.now + timedelta(days=1)),
            method="/Fred.Registry.Api.Registrar.Admin/add_registrar_zone_access",
            timeout=None,
        )

    async def test_update_zone_access_noop(self):
        self.client.mock.return_value = make_awaitable(self._get_update_zone_access_reply())

        self.assertEqual(
            await self.client.update_zone_access("REG-SW", ZoneAccess(zone="cz", valid_from=self.now)),
            ZoneAccess(zone="cz", valid_from=self.now),
        )

        request = UpdateRegistrarZoneAccessRequest()
        request.registrar_handle.value = "REG-SW"
        request.zone = "cz"
        request.valid_from.FromDatetime(self.now)
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Registrar.Admin/update_registrar_zone_access",
            timeout=None,
        )

    async def test_update_zone_access_null(self):
        self.client.mock.return_value = make_awaitable(self._get_update_zone_access_reply())

        self.assertEqual(
            await self.client.update_zone_access(
                "REG-SW",
                ZoneAccess(zone="cz", valid_from=self.tomorrow),
                valid_from=None,
                valid_to=None,
            ),
            ZoneAccess(zone="cz", valid_from=self.now),
        )

        request = UpdateRegistrarZoneAccessRequest()
        request.registrar_handle.value = "REG-SW"
        request.zone = "cz"
        request.valid_from.FromDatetime(self.tomorrow)
        request.set_valid_from.SetInParent()
        request.set_valid_to.SetInParent()
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Registrar.Admin/update_registrar_zone_access",
            timeout=None,
        )

    async def test_update_zone_access_datetime(self):
        self.client.mock.return_value = make_awaitable(self._get_update_zone_access_reply(valid_to=self.tomorrow))

        self.assertEqual(
            await self.client.update_zone_access(
                "REG-SW",
                ZoneAccess(zone="cz", valid_from=datetime(2020, 1, 1, tzinfo=timezone.utc)),
                valid_from=self.now,
                valid_to=self.tomorrow,
            ),
            ZoneAccess(zone="cz", valid_from=self.now, valid_to=self.tomorrow),
        )

        request = UpdateRegistrarZoneAccessRequest()
        request.registrar_handle.value = "REG-SW"
        request.zone = "cz"
        request.valid_from.FromDatetime(datetime(2020, 1, 1, tzinfo=timezone.utc))
        request.set_valid_from.FromDatetime(self.now)
        request.set_valid_to.FromDatetime(self.tomorrow)
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Registrar.Admin/update_registrar_zone_access",
            timeout=None,
        )

    async def test_delete_zone_access(self):
        reply = DeleteRegistrarZoneAccessReply()
        self.client.mock.return_value = make_awaitable(reply)

        await self.client.delete_zone_access("REG-SW", ZoneAccess(zone="cz", valid_from=self.now))

        request = DeleteRegistrarZoneAccessRequest()
        request.registrar_handle.value = "REG-SW"
        request.zone = "cz"
        request.valid_from.FromDatetime(self.now)
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Registrar.Admin/delete_registrar_zone_access",
            timeout=None,
        )
